@@ -1,15 +1,13 @@
 import os
-from multiprocessing import Process, Pool
+import rpyc  # type: ignore
 from shutil import rmtree
 import sqlite3
-from pathlib import Path
 import json
-from typing import List, Tuple
+from multiprocessing import Pool
+from pathlib import Path
 import fire  # type: ignore
 from tqdm import tqdm  # type: ignore
-import filetype  # type: ignore
 from .persistor import Persistor
-from .daemon import start
 from .graph import graph
 
 
@@ -21,39 +19,76 @@ class HyperTag:
         self.root_dir = Path(self.db.get_hypertagfs_dir())
         os.makedirs(self.root_dir, exist_ok=True)
 
-    def get_text_documents(self, file_paths: List[str]) -> List[Tuple[str, str]]:
-        doc_id = self.db.get_tag_id_by_name("Documents")
-        doc_types = set()
-        for tag_id, _type_name in self.db.get_tag_id_children_ids_names(doc_id):
-            doc_types.add(self.db.get_tag_name_by_id(tag_id))
+    def search_image(self, *text_queries: str, top_k=10, path=False, score=False, _return=False):
+        """ Execute a semantic search that returns best matching images """
+        text_query = " ".join(text_queries)
+        try:
+            rpc = rpyc.connect("localhost", 18861)
+            results = rpc.root.search_image(text_query, path, top_k, score)
+            for result in results:
+                print(result)
+            if len(result) == 0:
+                print("No relevant files indexed...")
+        except ConnectionRefusedError:
+            from .vectorizer import CLIPVectorizer
 
-        # Keep only text files
-        compatible_files = []
-        for file_path in file_paths:
-            file_type_guess = filetype.guess(str(file_path))
-            if file_type_guess is None:
-                continue
-            file_type = file_type_guess.extension.lower()
-            if file_type in doc_types:
-                compatible_files.append((str(file_path), file_type))
-        return compatible_files
+            vectorizer = CLIPVectorizer()
+            results = vectorizer.search_image(text_query, path, top_k, score)
+        if _return:
+            return results
 
-    def index(self, rebuild=False, cache=False, cores: int = 0):
+    def index(self, text=None, image=None, rebuild=False, cache=False, cores: int = 0):
+        """ Vectorize image & text files (needed for semantic search) """
+        if (image and text is None) or (image and text) or (not image and not text):
+            self.index_images()
+        if (text and image is None) or (image and text) or (not image and not text):
+            self.index_texts(rebuild, cache, cores)
+
+    def index_images(self):
+        """ Vectorize image files (needed for semantic search) """
+        from .vectorizer import CLIPVectorizer, get_image_files
+
+        file_paths = self.db.get_unindexed_file_paths()
+        compatible_files = get_image_files(file_paths, verbose=True)
+        print("Vectorizing", len(compatible_files), "images...")
+        remote = True
+        if remote:
+            try:
+                rpc = rpyc.connect("localhost", 18861)
+                rpc._config["sync_request_timeout"] = None  # Disable timeout
+                print("Connected to DaemonService successfully")
+            except ConnectionRefusedError:
+                print("DaemonService connection failed, falling back to local execution...")
+                remote = False
+
+        if not remote:
+            img_vectorizer = CLIPVectorizer()
+
+        for file_path in tqdm(compatible_files):
+            if remote:
+                img_vector = json.loads(rpc.root.encode_image(file_path))[0]
+            else:
+                img_vector = img_vectorizer.encode_image(file_path)[0].tolist()
+
+            self.db.add_file_embedding_vector(file_path, json.dumps(img_vector))
+            self.db.conn.commit()
+
+    def index_texts(self, rebuild=False, cache=False, cores: int = 0):
         """ Vectorize text files (needed for semantic search) """
         # TODO: index images
         # TODO: auto index on file addition (import)
-        print("Vectorizing text documents...")
-        from .vectorizer import (
-            extract_clean_text,
-            compute_text_embedding,
-        )
-        import torch
+        from .vectorizer import TextVectorizer, extract_clean_text, get_text_documents
 
-        cuda = torch.cuda.is_available()
-        if cuda:
-            print("Using CUDA to speed stuff up")
-        else:
-            print("CUDA not available (this might take a while)")
+        print("Vectorizing text documents...")
+        remote = True
+        if remote:
+            try:
+                rpc = rpyc.connect("localhost", 18861)
+                rpc._config["sync_request_timeout"] = None  # Disable timeout
+                print("Connected to DaemonService successfully")
+            except ConnectionRefusedError:
+                print("DaemonService connection failed, falling back to local execution...")
+                remote = False
         if cache:
             print("Caching cleaned texts (database will grow big)")
         if rebuild:
@@ -62,7 +97,7 @@ class HyperTag:
         else:
             file_paths = self.db.get_unindexed_file_paths()
         i = 0
-        compatible_files = self.get_text_documents(file_paths)
+        compatible_files = get_text_documents(file_paths, verbose=True)
         min_words = 5
         min_word_length = 4
         args = []
@@ -82,11 +117,16 @@ class HyperTag:
                 t.update(1)
                 if sentences:
                     inference_tuples.append((file_path, sentences))
-        print(f"Cleaned {len(inference_tuples)} text docs successfully")
+        print(f"Cleaned {len(inference_tuples)} text doc/s successfully")
         print("Starting inference...")
         # Compute embeddings
+        if not remote:
+            vectorizer = TextVectorizer()
         for file_path, sentences in tqdm(inference_tuples):
-            document_vector = compute_text_embedding(sentences)
+            if remote:
+                document_vector = json.loads(rpc.root.compute_text_embedding(json.dumps(sentences)))
+            else:
+                document_vector = vectorizer.compute_text_embedding(sentences)
             if (
                 document_vector is not None
                 and type(document_vector) is list
@@ -97,42 +137,28 @@ class HyperTag:
                 i += 1
             else:
                 print(type(document_vector))
+                self.db.add_file_embedding_vector(file_path, json.dumps([]))
+                self.db.conn.commit()
                 print("Failed to parse file - skipping:", file_path)
         print(f"Vectorized {str(i)} file/s successfully")
 
-    def search(self, text_query: str, path=False, top_k=10, score=False):
+    def search(self, *text_queries: str, path=False, top_k=10, score=False, _return=False):
         """ Execute a semantic search that returns best matching text documents """
-        from .vectorizer import (
-            clean_transform,
-            compute_text_embedding,
-            vector_search,
-        )
+        text_query = " . ".join(text_queries)
+        try:
+            rpc = rpyc.connect("localhost", 18861)
+            results = rpc.root.search(text_query, path, top_k, score)
+            for result in results:
+                print(result)
+            if len(result) == 0:
+                print("No relevant files indexed...")
+        except ConnectionRefusedError:
+            from .vectorizer import TextVectorizer
 
-        text_document_tuples = self.get_text_documents(self.db.get_files(show_path=True))
-        text_document_paths = [path for path, _file_type in text_document_tuples]
-        corpus = self.db.get_file_embedding_vectors(text_document_paths)
-        if len(corpus) == 0:
-            print("No relevant files indexed...")
-            return
-        sentence_query = clean_transform(text_query, 0, 1)
-        query_vector = compute_text_embedding(sentence_query)
-        corpus_paths = []
-        corpus_vectors = []
-        for doc_path, embedding_vector in corpus:
-            corpus_vectors.append(json.loads(embedding_vector))
-            corpus_paths.append(doc_path)
-
-        top_matches = vector_search(query_vector, corpus_vectors, top_k=top_k)
-        for match in top_matches[0]:
-            corpus_id, score_value = match["corpus_id"], match["score"]
-            file_path = corpus_paths[corpus_id]
-            file_name = file_path.split("/")[-1]
-            if path:
-                file_name = file_path
-            if score:
-                print(file_name, f"({score_value})")
-            else:
-                print(file_name)
+            vectorizer = TextVectorizer()
+            results = vectorizer.search(text_query, path, top_k, score)
+        if _return:
+            return results
 
     def set_hypertagfs_dir(self, path: str):
         """ Set path for HyperTagFS directory """
@@ -140,17 +166,21 @@ class HyperTag:
 
     def mount(self, root_dir=None, parent_tag_id=None):
         """ Generate HyperTagFS: tag representation using symlinks """
+        if root_dir is None:
+            root_dir = self.root_dir
+        root_path = Path(root_dir)
+
         if parent_tag_id is None:
             print("Updating HyperTagFS...")
             graph()
+            os.makedirs(root_path / "Search Texts", exist_ok=True)
+            os.makedirs(root_path / "Search Images", exist_ok=True)
             tag_ids_names = self.db.get_root_tag_ids_names()
         else:
             tag_ids_names = self.db.get_tag_id_children_ids_names(parent_tag_id)
 
         leaf_tag_ids = {tag_id[0] for tag_id in self.db.get_leaf_tag_ids()}
-        if root_dir is None:
-            root_dir = self.root_dir
-        root_path = Path(root_dir)
+
         for tag_id, name in tag_ids_names:
             file_paths_names = self.db.get_file_paths_names_by_tag_id(tag_id)
             root_tag_path = root_path / name
@@ -182,10 +212,10 @@ class HyperTag:
             if not is_hidden:
                 visible_file_paths.append(p)
         print("Adding files...")
-        self.add(*visible_file_paths)
+        added_file_paths = self.add(*visible_file_paths)
         import_path_dirs = set(str(import_path).split("/"))
         print("Adding tags...")
-        for file_path in tqdm(visible_file_paths):
+        for file_path in tqdm(added_file_paths):
             file_path_tags = [p for p in str(file_path).split("/") if p not in import_path_dirs][
                 :-1
             ]
@@ -205,16 +235,17 @@ class HyperTag:
 
     def add(self, *file_paths):
         """ Add file/s """
-        added = 0
+        added = []
         for file_path in tqdm(file_paths):
             try:
                 if Path(file_path).is_file():
                     self.db.add_file(os.path.abspath(file_path))
-                    added += 1
+                    added.append(file_path)
             except sqlite3.IntegrityError:
                 pass
         self.db.conn.commit()
-        print("Added", added, "new file/s")
+        print("Added", len(added), "new file/s")
+        return added
 
     def tags(self, *file_names):
         """ Display all tags of file/s """
@@ -376,6 +407,11 @@ class HyperTag:
 
 def daemon():
     """ Start daemon process watching HyperTagFS """
+    print("Starting up daemon...")
+    from .daemon import start
+    from multiprocessing import Process, set_start_method
+
+    set_start_method("spawn")
     p = Process(target=start)
     p.start()
     p.join()
@@ -412,6 +448,7 @@ def main():
         "daemon": daemon,
         "graph": graph,
         "index": ht.index,
+        "search_image": ht.search_image,
         "search": ht.search,
     }
     fire.Fire(fire_cli)
